﻿using System;
using System.Diagnostics;
using System.Drawing;

using Emgu.CV;
using Emgu.CV.CvEnum;
using Emgu.CV.Features2D;
using Emgu.CV.Structure;
using Emgu.CV.Util;
#if !IOS
using Emgu.CV.GPU;
#endif

namespace TestProject
{
    /// <summary>
    /// 매칭 헬퍼
    /// </summary>
    public static class MatchHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 매칭 찾기 - FindMatch(modelImage, sourceImage, matchTime, modelVectorOfKeyPoint, sourceVectorOfKeyPoint, indexMatrix, maskMatrix, homographyMatrix)

        /// <summary>
        /// 매칭 찾기
        /// </summary>
        /// <param name="modelImage">모델 이미지</param>
        /// <param name="sourceImage">소스 이미지</param>
        /// <param name="matchTime">매칭 시간</param>
        /// <param name="modelVectorOfKeyPoint">모델 키 포인트 벡터</param>
        /// <param name="sourceVectorOfKeyPoint">소스 키 포인트 벡터</param>
        /// <param name="indexMatrix">인덱스 매트릭스</param>
        /// <param name="maskMatrix">마스크 매트릭스</param>
        /// <param name="homographyMatrix">호모 그래피 매트릭스</param>
        public static void FindMatch(Image<Gray, Byte> modelImage, Image<Gray, byte> sourceImage, out long matchTime, out VectorOfKeyPoint modelVectorOfKeyPoint, out VectorOfKeyPoint sourceVectorOfKeyPoint, out Matrix<int> indexMatrix, out Matrix<byte> maskMatrix, out HomographyMatrix homographyMatrix)
        {
            int k = 2;

            double uniquenessThreshold = 0.8;

            SURFDetector surfDetector = new SURFDetector(500, false);

            Stopwatch stopwatch;

            homographyMatrix = null;

#if !IOS
            if(GpuInvoke.HasCuda)
            {
                GpuSURFDetector gpuSURFDetector = new GpuSURFDetector(surfDetector.SURFParams, 0.01f);

                using(GpuImage<Gray, Byte> modelGpuImage = new GpuImage<Gray, byte>(modelImage))
                {
                    using(GpuMat<float> modelKeyPointGpuMat = gpuSURFDetector.DetectKeyPointsRaw(modelGpuImage, null))
                    {
                        using(GpuMat<float> modelDescriptorGpuMat = gpuSURFDetector.ComputeDescriptorsRaw(modelGpuImage, null, modelKeyPointGpuMat))
                        {
                            using(GpuBruteForceMatcher<float> matcher = new GpuBruteForceMatcher<float>(DistanceType.L2))
                            {
                                modelVectorOfKeyPoint = new VectorOfKeyPoint();

                                gpuSURFDetector.DownloadKeypoints(modelKeyPointGpuMat, modelVectorOfKeyPoint);

                                stopwatch = Stopwatch.StartNew();

                                using(GpuImage<Gray, Byte> sourceGpuImage = new GpuImage<Gray, byte>(sourceImage))
                                {
                                    using(GpuMat<float> sourceKeyPointGpuMat = gpuSURFDetector.DetectKeyPointsRaw(sourceGpuImage, null))
                                    {
                                        using(GpuMat<float> sourceDescriptorGpuMat = gpuSURFDetector.ComputeDescriptorsRaw(sourceGpuImage, null, sourceKeyPointGpuMat))
                                        {
                                            using(GpuMat<int> matchIndexGpuMat = new GpuMat<int>(sourceDescriptorGpuMat.Size.Height, k, 1, true))
                                            {
                                                using(GpuMat<float> matchDistainceGpuMat = new GpuMat<float>(sourceDescriptorGpuMat.Size.Height, k, 1, true))
                                                {
                                                    using(GpuMat<Byte> maskGpuMat = new GpuMat<byte>(matchIndexGpuMat.Size.Height, 1, 1))
                                                    {
                                                        using(Stream stream = new Stream())
                                                        {
                                                            matcher.KnnMatchSingle(sourceDescriptorGpuMat, modelDescriptorGpuMat, matchIndexGpuMat, matchDistainceGpuMat, k, null, stream);

                                                            indexMatrix = new Matrix<int>(matchIndexGpuMat.Size);

                                                            maskMatrix = new Matrix<byte>(maskGpuMat.Size);

                                                            using(GpuMat<float> col0 = matchDistainceGpuMat.Col(0))
                                                            {
                                                                using(GpuMat<float> col1 = matchDistainceGpuMat.Col(1))
                                                                {
                                                                    GpuInvoke.Multiply(col1, new MCvScalar(uniquenessThreshold), col1, stream);

                                                                    GpuInvoke.Compare(col0, col1, maskGpuMat, CMP_TYPE.CV_CMP_LE, stream);
                                                                }
                                                            }

                                                            sourceVectorOfKeyPoint = new VectorOfKeyPoint();

                                                            gpuSURFDetector.DownloadKeypoints(sourceKeyPointGpuMat, sourceVectorOfKeyPoint);

                                                            stream.WaitForCompletion();

                                                            maskGpuMat.Download(maskMatrix);

                                                            matchIndexGpuMat.Download(indexMatrix);

                                                            if(GpuInvoke.CountNonZero(maskGpuMat) >= 4)
                                                            {
                                                                int nonZeroCount = Features2DToolbox.VoteForSizeAndOrientation
                                                                (
                                                                    modelVectorOfKeyPoint,
                                                                    sourceVectorOfKeyPoint,
                                                                    indexMatrix,
                                                                    maskMatrix,
                                                                    1.5,
                                                                    20
                                                                );

                                                                if(nonZeroCount >= 4)
                                                                {
                                                                    homographyMatrix = Features2DToolbox.GetHomographyMatrixFromMatchedFeatures
                                                                    (
                                                                        modelVectorOfKeyPoint,
                                                                        sourceVectorOfKeyPoint,
                                                                        indexMatrix,
                                                                        maskMatrix,
                                                                        2
                                                                    );
                                                                }
                                                            }

                                                            stopwatch.Stop();
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else
#endif
            {
                modelVectorOfKeyPoint = new VectorOfKeyPoint();

                Matrix<float> modelDescriptorMatrix = surfDetector.DetectAndCompute(modelImage, null, modelVectorOfKeyPoint);

                stopwatch = Stopwatch.StartNew();

                sourceVectorOfKeyPoint = new VectorOfKeyPoint();

                Matrix<float> sourceDescriptorMatrix = surfDetector.DetectAndCompute(sourceImage, null, sourceVectorOfKeyPoint);

                BruteForceMatcher<float> matcher = new BruteForceMatcher<float>(DistanceType.L2);

                matcher.Add(modelDescriptorMatrix);

                indexMatrix = new Matrix<int>(sourceDescriptorMatrix.Rows, k);

                using(Matrix<float> distanceMatrix = new Matrix<float>(sourceDescriptorMatrix.Rows, k))
                {
                    matcher.KnnMatch(sourceDescriptorMatrix, indexMatrix, distanceMatrix, k, null);

                    maskMatrix = new Matrix<byte>(distanceMatrix.Rows, 1);

                    maskMatrix.SetValue(255);

                    Features2DToolbox.VoteForUniqueness(distanceMatrix, uniquenessThreshold, maskMatrix);
                }

                int nonZeroCount = CvInvoke.cvCountNonZero(maskMatrix);

                if(nonZeroCount >= 4)
                {
                    nonZeroCount = Features2DToolbox.VoteForSizeAndOrientation
                    (
                        modelVectorOfKeyPoint,
                        sourceVectorOfKeyPoint,
                        indexMatrix,
                        maskMatrix,
                        1.5,
                        20
                    );

                    if(nonZeroCount >= 4)
                    {
                        homographyMatrix = Features2DToolbox.GetHomographyMatrixFromMatchedFeatures
                        (
                            modelVectorOfKeyPoint,
                            sourceVectorOfKeyPoint,
                            indexMatrix,
                            maskMatrix,
                            2
                        );
                    }
                }

                stopwatch.Stop();
            }

            matchTime = stopwatch.ElapsedMilliseconds;
        }

        #endregion
        #region 그리기 - Draw(modelImage, sourceImage, matchTime)

        /// <summary>
        /// 그리기
        /// </summary>
        /// <param name="modelImage">모델 이미지</param>
        /// <param name="sourceImage">소스 이미지</param>
        /// <param name="matchTime">매칭 시간</param>
        /// <returns>이미지</returns>
        public static Image<Bgr, Byte> Draw(Image<Gray, Byte> modelImage, Image<Gray, byte> sourceImage, out long matchTime)
        {
            HomographyMatrix homographyMatrix;
            VectorOfKeyPoint modelVectorOfKeyPoint;
            VectorOfKeyPoint sourceVectorOfKeyPoint;
            Matrix<int>      indexMatrix;
            Matrix<byte>     maskMatrix;

            FindMatch
            (
                modelImage,
                sourceImage,
                out matchTime,
                out modelVectorOfKeyPoint,
                out sourceVectorOfKeyPoint,
                out indexMatrix,
                out maskMatrix,
                out homographyMatrix
            );

            Image<Bgr, Byte> resultImage = Features2DToolbox.DrawMatches
            (
                modelImage,
                modelVectorOfKeyPoint,
                sourceImage,
                sourceVectorOfKeyPoint,
                indexMatrix,
                new Bgr(255, 255, 255),
                new Bgr(255, 255, 255),
                maskMatrix,
                Features2DToolbox.KeypointDrawType.DEFAULT
            );

            if(homographyMatrix != null)
            {
                Rectangle rectangle = modelImage.ROI;

                PointF[] pointArray = new PointF[]
                { 
                    new PointF(rectangle.Left , rectangle.Bottom),
                    new PointF(rectangle.Right, rectangle.Bottom),
                    new PointF(rectangle.Right, rectangle.Top   ),
                    new PointF(rectangle.Left , rectangle.Top   )
                };

                homographyMatrix.ProjectPoints(pointArray);

                resultImage.DrawPolyline
                (
                    Array.ConvertAll<PointF, Point>(pointArray, Point.Round),
                    true,
                    new Bgr(Color.Red),
                    5
                );
            }

            return resultImage;
        }

        #endregion
    }
}